class JSONDOM {
    constructor({ name, element }) {
        this.name = name || 'Element';
        this.element = element || null;
        this.children = [];
        this.props = {};

        return new Proxy(this, {
            get: (target, prop) => {
                if (prop in target) return target[prop];
                if (prop in target.props) return target.props[prop];
                target.props[prop] = null;
                return null;
            },
            set: (target, prop, value) => {
                if (prop in target) {
                    target[prop] = value;
                } else {
                    target.props[prop] = value;
                }
                return true;
            },
            ownKeys: (target) => [...Object.keys(target), ...Object.keys(target.props)],
            getOwnPropertyDescriptor: (target, prop) => {
                if (prop in target) return Reflect.getOwnPropertyDescriptor(target, prop);
                if (prop in target.props) {
                    return { configurable: true, enumerable: true, value: target.props[prop] };
                };
                return undefined;
            }
        });
    }
};

// -------------------------------

const json_to_json_dom = ({ structure }) => {
	const { name, element, children, ...props } = structure; 
	const result = new JSONDOM({ name: name, element: element });
	const child = Array.isArray(children) ? children.map((child) => { return json_to_json_dom(child) }) : children;
	result.children = child;
	return Object.assign( result, props );
};

const dom_to_json = ({ structure }) => {
	return new DOMParser().parseFromString(structure, "application/xml");;
};

const dom_to_json_dom = (structure) => {
	return json_to_json_dom(dom_to_json(structure));
};

const Indentify = (structure) => {
	const types = [
		{ type: 'JSON', code: 1 },
		{ type: 'DOM', code: 2 }
	];
	if (structure instanceof JSONDOM || structure[0] instanceof JSONDOM) {
		return types[0];
	} else if (structure instanceof HTMLElement || structure.$$typeof === Symbol.for('react.element')) {
		return types[1];
	}; 
	try {
		const json = JSON.parse(structure);
		const result = (json?.element) ? types[0] : types[1];
		return result;
	} catch (e) {
		return 0; 
	};
};

const isParseable = (value) => {
	if (value) {
		return true;
	};
	throw new Error(`Invalid value: DOM, Valid JSON, or JSON-DOM expected, found: ${typeof value}`);
};

const Tag = ({ element: Element, props, children }) => {
    return (
        <Element {...props}>
            {children}
        </Element>
    );
};

const SelfClosingTag = ({ element: Element, props }) => {
    return <Element {...props} />;
};

const json_to_dom = ({ structure }) => {
	return structure.map((item, index) => {
        const { selfClosing, _props, name, element, children, ...props } = item;

        const TagType = typeof element === 'string' ? (selfClosing ? SelfClosingTag : Tag) : element;

        const tagParams = { element, props };

        if (!selfClosing && children) {
            if (Array.isArray(children)) {
                tagParams.children = json_to_dom({ structure: children });
            } else {
                tagParams.children = children;
            };
        };

        return <TagType key={index} {...tagParams} />;
    });
};

const json_dom_to_json = ({ structure }) => {
	/*
		-- Fix Warning
		Warning: React does not recognize the `toJSON` prop on a DOM element. If you intentionally want it to appear in the DOM as a custom attribute, spell it as lowercase `tojson` instead. If you accidentally passed it from a parent component, remove it from the DOM element.
		div 
		
		delete structure.toJSON;
		delete structure._props.toJSON;
	*/
	return JSON.parse(JSON.stringify(structure));
};

// -------------------------------

const onParseToJsonWarning  = () => {
	console.log('Warning: Using JSON-DOM is recommended in favor of plain JSON Objects, if you really need to use plain JSON Objects you can disable this warning by setting: { warning: false } when declaring new Parse.');
};

class Parse {
	// -- Fix Chaining
	constructor({ structure, warning = true }) {
		this.value = Array.isArray(structure) ? structure[0] : structure;
		const indentified = Indentify(this.value);
		this.type = isParseable(indentified) && indentified.type;
		this.typeCode = indentified.code;
		this.warning = warning;
	};
	JSON() {
		if (this.typeCode === 1) {
			this.value = json_dom_to_json({ structure: [this.value] });
		} else if (this.typeCode === 2) {
			this.value = dom_to_json({ structure: this.value });
		};
		if (this.warning) {
			onParseToJsonWarning();
		};
		return this;
	};
	JSONDOM() {
		if (this.typeCode === 1) {
			this.value = json_to_json_dom({ structure: this.value });
		} else if (this.typeCode === 2) {
			this.value = dom_to_json_dom({ structure: this.value });
		};
		return this;
	};
	DOM() {
		this.value = json_to_dom({ structure: [this.value] });
		return this;
    };
};

export { JSONDOM, Parse };
